/*
 * TCPHA       An implementation of the TCP Virtual Server daemon inside
 *              kernel for the LINUX operating system.Different from KTCPVS,
 *		  TCPHA implements kenel layer-7 switch based on TCP Handoff.
 *		  TCPHA can be used to build a high-performance and highly available 
 *		   server based on a cluster of Linux servers(currently only 
 *              support BE which installs Linux,but i think it is NOT difficult to let Windows 
 *		   support TCP Handoff:),we will explore)
 *
 * Authors:     Li Wang  <dragonfly@linux-vs.org>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 */

#define __KERNEL_SYSCALLS__	/*  for waitpid */

#include "tcpha_be.h"
#include "tcpha_be_file.h"
#include "tcpha_be_hash.h"
#include "tcpha_be_misc.h"
#include "tcpha_be_con.h"

enum {
	SERVER_DEAD = 0,
	SERVER_STARTING,
	SERVER_READY,
	SERVER_BUSY
};
	
#define TCPHA_BE_MAX_DAEMONS		128
#ifndef MAX_SPAWN_RATE
#define MAX_SPAWN_RATE	32
#endif

struct tcpha_be_workshop {
	struct tcpha_be_worker workers[TCPHA_BE_MAX_DAEMONS];
	int max_daemons_limit;
	int idle_spawn_rate;
	unsigned long last_modified;	/* last time of add/killing child */
};

#define NET_KTCPHABE			20
#define NET_KTCPHABE_UNLOAD	2

static int errno;
atomic_t tcphabe_adv_worker_count;
int advwkrpid;
int tcphabe_unload = 0;

static struct ctl_table_header *ktcphabe_table_header;

static ctl_table ktcphabe_table[] = {
	{NET_KTCPHABE_UNLOAD, "unload", &tcphabe_unload,
	 sizeof(int), 0644, NULL, &proc_dointvec},
	{0}
};

static ctl_table ktcphabe_dir_table[] = {
	{NET_KTCPHABE, "ktcphabe", NULL, 0, 0555, ktcphabe_table},
	{0}
};

static ctl_table ktcphabe_root_table[] = {
	{CTL_NET, "net", NULL, 0, 0555, ktcphabe_dir_table},
	{0}
};

static inline void
update_worker_status(struct tcpha_be_worker *worker, __u16 status)
{
	worker->status = status;
}

static int worker_daemon(void *);
static int adv_worker_daemon(void *);

static inline void
hire_worker(struct tcpha_be_workshop *tbl,
	   int slot, struct tcpha_be_service *svc)
{
	if (slot + 1 > tbl->max_daemons_limit)
		tbl->max_daemons_limit = slot + 1;
	tbl->last_modified = jiffies;
	tbl->workers[slot].svc = svc;
	if (kernel_thread(worker_daemon, &tbl->workers[slot],
		CLONE_VM | CLONE_FS | CLONE_FILES) < 0)
		ERR("spawn child failed\n");
}

static inline void
hire_adv_worker(struct tcpha_be_service *svc)
{
	if (!atomic_read(&tcphabe_adv_worker_count)) {
		if (kernel_thread(adv_worker_daemon, svc ,
			  	CLONE_VM | CLONE_FS | CLONE_FILES) < 0)
			ERR("spawn child failed\n");
		else
			atomic_inc(&tcphabe_adv_worker_count);
	}
}

static inline void
fire_worker(struct tcpha_be_workshop *tbl, int slot,int cmd)
{
	kill_proc(tbl->workers[slot].pid,cmd, 1);
	tbl->last_modified = jiffies;
}

static inline void
tell_workers_offduty(struct tcpha_be_workshop *tbl)
{
	int i;

	for (i=0;i<tbl->max_daemons_limit;i++) {
		if (tbl->workers[i].status!=SERVER_DEAD) {
			fire_worker(tbl,i,SIGSTOP);
		}
	}
	if (atomic_read(&tcphabe_adv_worker_count))
		kill_proc(advwkrpid,SIGSTOP,1);
}

static inline void
workshop_maintenance(struct tcpha_be_workshop *tbl,
		       struct tcpha_be_service *svc)
{
	int i;
	int free_slots[MAX_SPAWN_RATE];
	int free_length = 0;
	int to_kill = -1;
	int idle_count = 0;
	int last_non_dead = -1;

	hire_adv_worker(svc);
	for (i = 0; i < svc->conf.maxClients; i++) {
		int status;

		if (i >= tbl->max_daemons_limit
		    && free_length == tbl->idle_spawn_rate) break;
		status = tbl->workers[i].status;
		switch (status) {
		case SERVER_DEAD:
			if (free_length < tbl->idle_spawn_rate) {
				free_slots[free_length] = i;
				free_length++;
			}
			break;
		case SERVER_STARTING:
			idle_count++;
			last_non_dead = i;
			break;
		case SERVER_READY:
			idle_count++;
			to_kill = i;
			last_non_dead = i;
			break;
		case SERVER_BUSY:
			last_non_dead = i;
			break;
		}
	}
	tbl->max_daemons_limit = last_non_dead + 1;

	if (idle_count > svc->conf.maxSpareServers) {
		/* fire one worker each time */
		fire_worker(tbl, to_kill,SIGKILL);
		tbl->idle_spawn_rate = 1;
	} else if (idle_count < svc->conf.minSpareServers) {
		if (free_length) {
			if (tbl->idle_spawn_rate > 8 && net_ratelimit())
				INFO
				    ("Server seems busy, you may "
				     "need to increase StartServers, "
				     "or Min/MaxSpareServers\n");
			/* spawn a batch of children */
			for (i = 0; i < free_length; i++)
				hire_worker(tbl, free_slots[i], svc);

			if (tbl->idle_spawn_rate < MAX_SPAWN_RATE)
				tbl->idle_spawn_rate *= 2;
		} else if (net_ratelimit())
			INFO
			    ("Server reached MaxClients setting, "
			     "consider raising the MaxClients "
			     "setting\n");
	} else {
		/* if the number of spare servers remains in the interval
		   (minSpareServers, maxSpareServers] and the time of
		   last modified is larger than ten minutes, we try to
		   kill one spare child in order to release some resource. */
		if (idle_count > svc->conf.minSpareServers
		    && jiffies - tbl->last_modified > 600 * HZ)
			fire_worker(tbl, to_kill,SIGKILL);
		tbl->idle_spawn_rate = 1;
	}
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
static inline void
__init_prc_seg(void)
{
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked,
		      sigmask(SIGKILL) | sigmask(SIGSTOP));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
}
#else
static inline void
__init_prc_seg(void)
{
	siginitsetinv(&current->blocked,
		      sigmask(SIGKILL) | sigmask(SIGSTOP));
	recalc_sigpending();
}
#endif
#define init_prc_seg() __init_prc_seg()

static int
worker_daemon(void *__wkr)
{
	struct tcpha_be_conn *conn;
	struct socket *sock;
	int ret = 0;
	struct tcpha_be_worker *worker = (struct tcpha_be_worker *) __wkr;
	struct tcpha_be_service *svc = worker->svc;

	DECLARE_WAITQUEUE(wait, current);

	atomic_inc(&svc->workercount);
	worker->pid = current->pid;
	update_worker_status(worker, SERVER_STARTING);

	sprintf(current->comm, "tcpha_be worker");
	lock_kernel();
	daemonize();

	/* Block all signals except SIGKILL and SIGSTOP */
	init_prc_seg();
	
	sock = svc->mainsock;
	if (sock == NULL) {
		ERR("socket for accept is NULL\n");
		ret = -1;
		goto out;
	}

	while (svc->stop == 0 && tcphabe_unload == 0) {
		if (signal_pending(current)) {
			DBG("worker (pid=%d): signal received\n",
				   current->pid);
			break;
		}

		update_worker_status(worker, SERVER_READY);
		if (sock->sk->tp_pinfo.af_tcp.accept_queue == NULL) {
			add_wait_queue(sock->sk->sleep, &wait);
			__set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ);
			__set_current_state(TASK_RUNNING);
			remove_wait_queue(sock->sk->sleep, &wait);
			continue;
		}
		conn = tcpha_be_conn_create(sock,svc);
		if (!conn)
			break;
		/* Do the actual accept */
		ret = sock->ops->accept(sock, conn->fsock, O_NONBLOCK);
		if (ret < 0) {
			tcpha_be_conn_release(conn);
			continue;
		}
		update_worker_status(worker, SERVER_BUSY);
		
		/* Do the work */
		ret = tcpha_be_conn_handle(conn, svc);
		tcpha_be_conn_release(conn);
		if (ret < 0) {
			DBG_RL("Error handling connection\n");
		}
		break;
	}

      out:
	update_worker_status(worker, SERVER_DEAD);
	atomic_dec(&svc->workercount);
	return ret;
}

static int
adv_worker_daemon(void *__svc)
{
	int ret = 0;
	char *Buffer;
	size_t BufLen;
	struct tcpha_be_service *svc = (struct tcpha_be_service *)__svc;
	

	atomic_inc(&svc->workercount);
	advwkrpid = current->pid;

	sprintf(current->comm, "tcpha_be advwkr");
	lock_kernel();
	daemonize();

	/* Block all signals except SIGKILL and SIGSTOP */
	init_prc_seg();

	Buffer = (char *) get_free_page(GFP_KERNEL);
	if (Buffer == NULL) {
		ret = -ENOMEM;
		goto out;
	}
	BufLen = PAGE_SIZE;
	
	while (svc->stop == 0 && tcphabe_unload == 0) {
		if (signal_pending(current)) {
			DBG("adv worker (pid=%d): signal received\n",
				   current->pid);
			break;
		}

		if  (skb_queue_empty(&pkt_buf.receive_queue)) {
			interruptible_sleep_on_timeout(&pkt_buf.wait, HZ);
			continue;
		}

		tcpha_be_adv_conn_handle(svc,Buffer,BufLen);
	}
	free_page((unsigned long) Buffer);

	out:
	atomic_dec(&tcphabe_adv_worker_count);
	atomic_dec(&svc->workercount);
	return ret;
}

static int
supervisor_daemon(void *__service)
{
	int waitpid_result;
	int i;
	struct tcpha_be_service *svc = (struct tcpha_be_service *) __service;
	struct tcpha_be_workshop *workshop = NULL;

	MOD_INC_USE_COUNT;
	
	DECLARE_WAIT_QUEUE_HEAD(WQ);

	sprintf(current->comm, "tcpha_be master");
	lock_kernel();
	daemonize();

	/* Block all signals except SIGKILL and SIGSTOP */
	init_prc_seg();
	
	workshop = vmalloc(sizeof(*workshop));
	if (!workshop)
		goto out;

	/* Then start listening and spawn the daemons */
	if (SetupServerSocket(svc) < 0)
		goto out;

	atomic_set(&svc->running, 1);
	atomic_set(&svc->workercount, 0);
	atomic_set(&tcphabe_adv_worker_count,0);
	svc->stop = 0;

	memset(workshop, 0, sizeof(*workshop));
	workshop->idle_spawn_rate = 1;

	hire_adv_worker(svc);
	for (i = 0; i < svc->conf.startservers; i++)
		hire_worker(workshop, i, svc);

	/* Then wait for deactivation */
	while (svc->stop == 0 && !signal_pending(current)
	       && tcphabe_unload == 0) {
		interruptible_sleep_on_timeout(&WQ, HZ);

		/* dynamically keep enough thread to handle load */
		workshop_maintenance(workshop, svc);

		/* reap the zombie daemons */
		waitpid_result = waitpid(-1, NULL, __WCLONE | WNOHANG);
	}

	tell_workers_offduty(workshop);
	
	/* Wait for tcp_vs_child to stop, one second per iteration */
	while (atomic_read(&svc->workercount) > 0)
		interruptible_sleep_on_timeout(&WQ, HZ);

	/* reap the zombie daemons */
	waitpid_result = 1;
	while (waitpid_result > 0)
		waitpid_result = waitpid(-1, NULL, __WCLONE | WNOHANG);

	/* stop listening */
	ReleaseServerSocket(svc);

      out:
	if (workshop)
		vfree(workshop);
	svc->start = 0;
	atomic_set(&svc->running, 0);

	MOD_DEC_USE_COUNT;
	
	return 0;
}

void init_pkt_buf(void)
{
	skb_queue_head_init(&pkt_buf.receive_queue);
	init_waitqueue_head(&pkt_buf.wait);
}

void flush_pkt_buf(void)
{
	struct sk_buff *skb;
	
	while (!skb_queue_empty(&pkt_buf.receive_queue)) {
		skb=__skb_dequeue(&pkt_buf.receive_queue);
		if (skb)
			kfree_skb(skb);
	}
}

void flush_be_list(struct tcpha_be_service *svc)
{
	struct list_head *l;
	struct tcpha_be *be;

	for (l = svc->be_list.next; l != &svc->be_list;) {
		be = list_entry(l, struct tcpha_be, list);
		l = be->list.next;
		list_del(&be->list);
		kfree(be);
	}
}

void flush_be_t(rule_t *rule)
{
	struct list_head *l;
	be_t *be;

	for (l = rule->be_list.next; l != &rule->be_list;) {
		be = list_entry(l, be_t, list);
		l = be->list.next;
		list_del(&be->list);
		kfree(be);
	}
}

void flush_rule_list(struct tcpha_be_service *svc)
{
	struct list_head *l;
	rule_t *rule;

	for (l = svc->rule_list.next; l != &svc->rule_list;) {
		rule = list_entry(l, rule_t, list);
		l = rule->list.next;
		list_del(&rule->list);
		flush_be_t(rule);
		regfree(&rule->rx);
		kfree(rule);
	}
}

void flush_tcpha_service(struct tcpha_be_service *svc)
{
	flush_be_list(svc);
	flush_rule_list(svc);
}

extern struct nf_hook_ops noarp_nf_ops_in;
extern struct nf_hook_ops noarp_nf_ops_out;
extern struct nf_hook_ops tcpha_be_in_ops;
extern struct nf_hook_ops tcpha_be_out_ops;

static int __init
tcpha_be_init(void)
{
	int ret=0;

	ret=config_tcpha_be(&tbs);
	if (ret < 0) {
		ERR("can't config tcpha service\n");
		goto out_error;
	}
	init_pkt_buf();
	ret = tcpha_be_connhash_init();
	if (ret < 0) {
		ERR("can't setup connection table.\n");
		goto out_error;
	}
       ret = nf_register_hook(&noarp_nf_ops_in);
	if (ret<0) {
		ERR("Register in arp hook error.\n");
		goto out_error;
	}
       ret = nf_register_hook(&noarp_nf_ops_out);
	if (ret<0) {
		ERR("Register out arp hook error.\n");
		goto out_nooutarphook;
	}
	ret = nf_register_hook(&tcpha_be_in_ops);
	if (ret < 0) {
		ERR("can't register in hook.\n");
		goto out_noinops;
	}
	ret = nf_register_hook(&tcpha_be_out_ops);
	if (ret < 0) {
		ERR("can't register out hook.\n");
		goto out_nooutops;
	}
	ktcphabe_table_header = register_sysctl_table(ktcphabe_root_table, 0);
	(void) kernel_thread(supervisor_daemon,&tbs, 0);
	INFO("loaded.\n");
	return ret;

	out_nooutops:
	nf_unregister_hook(&tcpha_be_in_ops);
	out_noinops:
	nf_unregister_hook(&noarp_nf_ops_out);
	out_nooutarphook:
       nf_unregister_hook(&noarp_nf_ops_in);
	out_error:
	return ret;
}


static void __exit
tcpha_be_cleanup(void)
{
	nf_unregister_hook(&tcpha_be_out_ops);
	nf_unregister_hook(&tcpha_be_in_ops);
	nf_unregister_hook(&noarp_nf_ops_in);
	nf_unregister_hook(&noarp_nf_ops_out);
	flush_pkt_buf();
	tcpha_be_conn_cleanup();
	unregister_sysctl_table(ktcphabe_table_header);
	flush_tcpha_service(&tbs);
	kmem_cache_destroy(tcpha_be_conn_cachep);
	INFO("unloaded.\n");
}

module_init(tcpha_be_init);
module_exit(tcpha_be_cleanup);
MODULE_LICENSE("GPL");
