/*
  *  
  *	see tcpha_be.c for infoes about TCPHA
  *
  */

#include "tcpha_be.h"
#include "tcpha_be_misc.h"
#include "tcpha_be_hash.h"
#include "tcpha_be_con.h"

#ifdef	CONFIG_TCPHA_BE_DEBUG
int tcpha_be_connmagic=0;
int tcpha_be_ctrlconnmagic=0;
#define tf_set_connmagic(x) 	x->magic=tcpha_be_connmagic++
#define tf_set_ctrlconnmagic(x)	x->magic=tcpha_be_ctrlconnmagic++;
#else	/* NO DEBUGGING at ALL */
#define tf_set_connmagic(x) do {} while (0)
#define tf_set_ctrlconnmagic(x)  do {} while (0)
#endif

static const char * state_name_table[TCPHA_BE_S_LAST+1] = {
	[TCPHA_BE_S_ESTABLISHED]	=	"ESTABLISHED",
	[TCPHA_BE_S_WILLHANDOFF] 	=	"WILLHANDOFF",
	[TCPHA_BE_S_HANDOFFING]		=	"HANDOFFING",
	[TCPHA_BE_S_HANDOFFED]		=	"HANDOFFED",
	[TCPHA_BE_S_TIME_WAIT]	=	"TIME_WAIT",
	[TCPHA_BE_S_LAST]          =	"BUG!",
};

/*
 *	Timeout table[state]
 */
struct timeout_table def_timeout_table = {
	ATOMIC_INIT(0),	/* refcnt */
	0,		/* scale  */
	{
		[TCPHA_BE_S_ESTABLISHED]	=	15*60*HZ,
		[TCPHA_BE_S_WILLHANDOFF] 	=     1*60*HZ,
		[TCPHA_BE_S_HANDOFFING]		=	1*60*HZ,
		[TCPHA_BE_S_HANDOFFED]		=	1*60*HZ,
		[TCPHA_BE_S_TIME_WAIT]	=	2*60*HZ,
		[TCPHA_BE_S_LAST]          =	2*HZ,
	},	/* timeout */
};

static struct timeout_table *tcpha_be_timeout_table = &def_timeout_table;

static inline void
tcpha_be_timeout_attach(struct tcpha_be_client_conn *cp, struct timeout_table *vstim)
{
	atomic_inc(&vstim->refcnt);
	cp->timeout_table = vstim;
}

int tcpha_be_set_state_timeout(struct tcpha_be_client_conn *cp, int state)
{
	struct timeout_table *vstim = cp->timeout_table;

	/*
	 *	Use default timeout table if no specific for this entry
	 */
	write_lock_bh(&cp->lock);
	if (!vstim)
		vstim = &def_timeout_table;

	cp->timeout = vstim->timeout[cp->state=state];

	if (vstim->scale) {
		int scale = vstim->scale;

		if (scale<0)
			cp->timeout >>= -scale;
		else if (scale > 0)
			cp->timeout <<= scale;
	}
	write_unlock_bh(&cp->lock);
	mod_timer(&cp->timer, jiffies+cp->timeout);
	
	return state;
}

static inline void tcpha_be_timeout_detach(struct tcpha_be_client_conn *cp)
{
	struct timeout_table *vstim = cp->timeout_table;

	if (!vstim)
		return;
	cp->timeout_table = NULL;
	atomic_dec(&vstim->refcnt);
}

static void tcpha_be_conn_expire(unsigned long data)
{
	struct tcpha_be_client_conn *cp = (struct tcpha_be_client_conn *)data;

	if (cp->timeout_table)
		cp->timeout = cp->timeout_table->timeout[TCPHA_BE_S_TIME_WAIT];
	else
		cp->timeout = def_timeout_table.timeout[TCPHA_BE_S_TIME_WAIT];

	/*
	 *	hey, I'm using it
	 */
	atomic_inc(&cp->refcnt);

	/*
	 *	unhash it if it is hashed in the conn table
	 */
	tcpha_be_conn_unhash(cp);

	/*
	 *	refcnt==1 implies I'm the only one referrer
	 */
	if (likely(atomic_read(&cp->refcnt) <= 1)) {
		/* make sure that there is no timer on it now */
		if (timer_pending(&cp->timer))
			del_timer(&cp->timer);

		tcpha_be_timeout_detach(cp);
		atomic_dec(&tcpha_be_conn_count);

		kmem_cache_free(tcpha_be_client_conn_cachep, cp);
		return;
	}

	/* hash it back to the table */
	tcpha_be_conn_hash(cp);

	DBG("delayed: refcnt-1=%d\n",
		  atomic_read(&cp->refcnt)-1);
	tcpha_be_conn_put(cp);
}


void tcpha_be_conn_expire_now(struct tcpha_be_client_conn *cp)
{
	cp->timeout = 0;
	mod_timer(&cp->timer, jiffies);
	__tcpha_be_conn_put(cp);
}

const char * tcpha_be_state_name(int state)
{
	if (state >= TCPHA_BE_S_LAST)
		return "ERR!";
	return state_name_table[state] ? state_name_table[state] : "?";
}

typedef enum ha_ret {
	TCP_HA_OK=0,
	TCP_HA_NO
}ha_ret_t;

typedef struct tcp_ha_ret {
	__u32 magic;
	__u32 connmagic;
	ha_ret_t msg;
}tcp_ha_ret_t;

struct tcpha_be_conn *
tcpha_be_conn_create(struct socket *sock, struct tcpha_be_service *svc)
{
	struct tcpha_be_conn *cp;

	cp = kmem_cache_alloc(tcpha_be_conn_cachep, GFP_KERNEL);
	if (!cp) {
		ERR_RL("create_conn no memory available\n");
		return NULL;
	}
	memset(cp, 0, sizeof(*cp));
	/* clone the socket */
	cp->fsock = sock_alloc();
	if (!cp->fsock) {
		kmem_cache_free(tcpha_be_conn_cachep, cp);
		return NULL;
	}
	tf_set_connmagic(cp);

	cp->fsock->type = sock->type;
	cp->fsock->ops = sock->ops;

	return cp;
}

static inline void 
tcpha_be_conn_getinfo(struct sock *sk,struct tcpha_conn_info *info)
{
	read_lock_irq(&sk->callback_lock);
	info->daddr=sk->daddr;
	info->saddr=sk->saddr;
	info->dport=sk->dport;
	info->sport=sk->sport;
	info->rcv_saddr=sk->rcv_saddr;
	info->af_tcp=sk->tp_pinfo.af_tcp;
	read_unlock_irq(&sk->callback_lock);
	info->connmagic=sk->tp_pinfo.af_tcp.rcv_nxt;
	info->magic=TCPHAMAGIC;
}

struct tcpha_be_client_conn *
tcpha_be_client_conn_create(void)
{
	struct tcpha_be_client_conn *cp;

	cp = kmem_cache_alloc(tcpha_be_client_conn_cachep, GFP_KERNEL);
	if (!cp) {
		ERR_RL("create_conn no memory available\n");
		return NULL;
	}
	memset(cp, 0, sizeof(*cp));
	
	init_timer(&cp->timer);
	cp->timer.data     = (unsigned long)cp;
	cp->timer.function = tcpha_be_conn_expire;
	tcpha_be_timeout_attach(cp, tcpha_be_timeout_table);
	rwlock_init(&cp->lock);
	INIT_LIST_HEAD(&cp->list);
	atomic_inc(&tcpha_be_conn_count);

	/* used by me */
	atomic_set(&cp->refcnt, 0);
	
	return cp;
}

static void tcp_reset(struct sock *sk)
{
	/* We want the right error as BSD sees it (and indeed as we do). */
	switch (sk->state) {
		case TCP_SYN_SENT:
			sk->err = ECONNREFUSED;
			break;
		case TCP_CLOSE_WAIT:
			sk->err = EPIPE;
			break;
		case TCP_CLOSE:
			return;
		default:
			sk->err = ECONNRESET;
	}

	if (!sk->dead)
		sk->error_report(sk);

	tcp_done(sk);
}

static __inline__ int tcp_hashfn(__u32 laddr, __u16 lport,
				 __u32 faddr, __u16 fport)
{
	int h = ((laddr ^ lport) ^ (faddr ^ fport));
	h ^= h>>16;
	h ^= h>>8;
	return h & (tcp_ehash_size - 1);
}

static inline struct sock *__tcp_v4_lookup_established(u32 saddr, u16 sport,
						       u32 daddr, u16 hnum, int dif)
{
	struct tcp_ehash_bucket *head;
	TCP_V4_ADDR_COOKIE(acookie, saddr, daddr)
	__u32 ports = TCP_COMBINED_PORTS(sport, hnum);
	struct sock *sk;
	int hash;

	/* Optimize here for direct hit, only listening connections can
	 * have wildcards anyways.
	 */
	hash = tcp_hashfn(daddr, hnum, saddr, sport);
	head = &tcp_ehash[hash];
	read_lock(&head->lock);
	for(sk = head->chain; sk; sk = sk->next) {
		if(TCP_IPV4_MATCH(sk, acookie, saddr, daddr, ports, dif))
			goto hit; /* You sunk my battleship! */
	}

	/* Must check for a TIME_WAIT'er before going to listener hash. */
	for(sk = (head + tcp_ehash_size)->chain; sk; sk = sk->next)
		if(TCP_IPV4_MATCH(sk, acookie, saddr, daddr, ports, dif))
			goto hit;
	read_unlock(&head->lock);

	return NULL;

hit:
	sock_hold(sk);
	read_unlock(&head->lock);
	return sk;
}

static inline struct sock *tcpha_be_conn_lookup(struct iphdr *iph,struct tcphdr *th)
{
	struct sock *sk=NULL;

	local_bh_disable();
	sk = __tcp_v4_lookup_established(iph->saddr, th->source,iph->daddr,ntohs(th->dest),0);
	if (sk&&sk->state!=TCP_ESTABLISHED) {
		sock_put(sk);
		sk=NULL;
	}
	local_bh_enable();
	if (!sk) {
		DBG("sk not found,or state not established\n");
	}
	return sk;
}

static inline struct sk_buff *tcpha_be_conn_getreq(struct socket *sock)
{
	struct sk_buff *skb;
	
	skb = skb_dequeue(&sock->sk->receive_queue);
	if (!skb) {
		goto out;
	}
	if (!pskb_may_pull(skb, TCPHACONNINFOSIZE)||!skb->data||*(__u32 *)skb->data!=TCPHAMAGIC) {
		kfree_skb(skb);
		skb=NULL;
		goto out;
	}
	/* we cannot release the skb here, but we do need to call
	   its destructor so that the sock_rfree can update the
	   source sk->rmem_alloc correctly. */
	if (skb->destructor) {
		skb->destructor(skb);
	}
	out:
	return skb;
}

static inline int tcpha_be_waitforpkt(struct socket *sock,long timeout)
{
	if (skb_queue_empty(&(sock->sk->receive_queue))) {
		if (!interruptible_sleep_on_timeout(&sock->wait, timeout)) {
			DBG("connection time out\n");
			goto out_error;
		}
		if (skb_queue_empty(&(sock->sk->receive_queue))) {
			DBG("connection closed by peer or received signal by process:%d\n",current->pid);
			goto out_error;
		}
	}
	return 0;
	out_error:
	return -1;
}

int 
tcpha_be_adv_conn_handle(struct tcpha_be_service *svc,char *buf,int buflen)
{
	struct sk_buff *skb;
	struct iphdr *iph;
	struct tcphdr *th;
	struct socket *dsock;
	struct tcpha_conn_info *info;
	struct tcpha_fe_info feinfo;
	struct sock *sk;
	int state;
	struct tcpha_be *be;
	tcp_ha_ret_t ret;
	char *buffer;
	int len;
	
	skb = skb_dequeue(&pkt_buf.receive_queue);
	if (!skb)
		return -1;
	iph=skb->nh.iph;
	th=skb->h.th;
	if (!iph||!th)
		goto out_error;
	be=(struct tcpha_be *)th->seq;
	DBG_RL("from client:%u.%u.%u.%u:%d,"
		"to:%u.%u.%u.%u:%d,"
		"scheduled BE:%u.%u.%u.%u:%d\n",
		NIPQUAD(iph->saddr),ntohs(th->source),
		NIPQUAD(iph->daddr),ntohs(th->dest),
		NIPQUAD(be->addr),ntohs(be->port));
	dsock=tcpha_be_connect2be(be->addr,be->port);
	if (!dsock) {
		DBG("connect BE:%u.%u.%u.%u:%d error\n",NIPQUAD(be->addr),ntohs(be->port));
		goto out_error;
	}
	buffer=buf+TCPHACONNINFOSIZE;
	len=buflen-TCPHACONNINFOSIZE;
	len=skb->len>len?len:skb->len;
	memcpy(buffer,skb->data,len);
	len+=TCPHACONNINFOSIZE;
	sk=tcpha_be_conn_lookup(iph,th);
	if (!sk) {
		goto out_release_error;
	}
	read_lock_bh(&((struct tcpha_be_client_conn *)th->ack_seq)->lock);
	state=((struct tcpha_be_client_conn *)th->ack_seq)->state;
	read_unlock_bh(&((struct tcpha_be_client_conn *)th->ack_seq)->lock);
	if (state!=TCPHA_BE_S_WILLHANDOFF) {
		goto out_close_error;
	}
	tcpha_be_set_state_timeout((struct tcpha_be_client_conn *)th->ack_seq,TCPHA_BE_S_HANDOFFING);
	info=(struct tcpha_conn_info *)buf;
	tcpha_be_conn_getinfo(sk,info);
	tcp_reset(sk);
	if (tcpha_be_sendbuffer(dsock, buf, len, 0) != len) {
		DBG_RL("send http buffer to %u.%u.%u.%u:%d error\n",
					NIPQUAD(be->addr),ntohs(be->port));
		goto out_close_error;
	}
	if (tcpha_be_waitforpkt(dsock,15*HZ))
		goto out_close_error;
	len = tcpha_be_recvbuffer(dsock,(char *)&ret,sizeof(ret),0);
	if (len!=sizeof(ret)||ret.connmagic!=info->connmagic||ret.magic!=info->magic) {
		DBG_RL("error reading handoff response from BE:%u.%u.%u.%u:%d\n",
					NIPQUAD(be->addr),ntohs(be->port));
		goto out_close_error;
	} 
	if (ret.msg!=TCP_HA_OK) {
		DBG_RL("handoff request refused by BE\n");
		goto out_close_error;
	}
	((struct tcpha_be_client_conn *)th->ack_seq)->ndaddr=be->addr;
	tcpha_be_set_state_timeout((struct tcpha_be_client_conn *)th->ack_seq, TCPHA_BE_S_HANDOFFED);
	sock_release(dsock);
	dsock=tcpha_be_connect2be(svc->conf.faddr,svc->conf.fport);
	if (!dsock) {
		DBG("connect to FE error\n");
		goto out_close_error;
	}
	feinfo.magic=TCPHAMAGIC;
	feinfo.caddr=iph->saddr;
	feinfo.cport=th->source;
	feinfo.daddr=iph->daddr;
	feinfo.dport=th->dest;
	feinfo.ndaddr=be->addr;
	if (tcpha_be_sendbuffer(dsock,(const char *)&feinfo,sizeof(feinfo), 0) != sizeof(feinfo)) {
		DBG_RL("send infoes to FE error\n");
		goto out_close_error;
	}
	if (tcpha_be_waitforpkt(dsock,15*HZ))
		goto out_close_error;
	len = tcpha_be_recvbuffer(dsock,(char *)&ret,sizeof(ret),0);
	if (len!=sizeof(ret)||ret.msg!=TCP_HA_OK||ret.magic!=feinfo.magic) {
		DBG_RL("error reading handoff response from FE\n");
		goto out_close_error;
	}
	tcpha_be_conn_expire_now((struct tcpha_be_client_conn *)th->ack_seq);
	sock_put(sk);
	sock_release(dsock);
	kfree_skb(skb);
	return 0;
	out_close_error:
	sock_put(sk);
	out_release_error:
	sock_release(dsock);
	out_error:
	kfree(skb);
	return -1;
}

int
redirect_to_local(struct tcpha_be_conn *,struct tcpha_conn_info *,struct sk_buff *skb, __u32, __u16);

int
tcpha_be_conn_handle(struct tcpha_be_conn *conn, struct tcpha_be_service *svc)
{
	struct socket *fsock;
	struct tcpha_conn_info *info;
	struct tcpha_be_client_conn *cp;
	struct sk_buff *skb;
	tcp_ha_ret_t ret;
	
	fsock= conn->fsock;
	if (fsock->sk->state != TCP_ESTABLISHED) {
		if (fsock->sk->state == TCP_CLOSE_WAIT) {
			goto out;
		}
		DBG_RL("Error control connection not established (state %d)\n",
			   fsock->sk->state);
		goto out_error;
	}
	DBG("server:%u.%u.%u.%u:%d created control connection:%d with %u.%u.%u.%u:%d by process:%d\n",
		       NIPQUAD(tbs.conf.raddr),ntohs(tbs.conf.port), conn->magic,
		       NIPQUAD(fsock->sk->daddr),ntohs(fsock->sk->dport),current->pid);
	again:
	if (tcpha_be_waitforpkt(fsock,15*60*HZ))
		goto out_error;
	if (!(skb=tcpha_be_conn_getreq(fsock))) {
		DBG_RL("error reading handoff request from %u.%u.%u.%u:%d\n",
					NIPQUAD(fsock->sk->daddr),ntohs(fsock->sk->dport));
		goto out_error;
	}
	info=(struct tcpha_conn_info *)skb->data;
	__skb_pull(skb,TCPHACONNINFOSIZE);
	ret.connmagic=info->connmagic;
	ret.msg=redirect_to_local(conn,info,skb,tbs.conf.redirect_addr,tbs.conf.redirect_port)?TCP_HA_NO:TCP_HA_OK;
	ret.magic=TCPHAMAGIC;
	if (tcpha_be_sendbuffer(fsock,(char *)&ret,sizeof(ret),0)!=sizeof(ret)) {
		DBG("send handoff response to %u.%u.%u.%u:%d error\n",
					NIPQUAD(fsock->sk->daddr),ntohs(fsock->sk->dport));
		goto out_error;
	}
	if (ret.msg==TCP_HA_OK) {
		cp=tcpha_be_client_conn_create();
		if (!cp)
			goto out_error;
		cp->caddr=info->daddr;
		cp->cport=info->dport;
		cp->daddr=info->saddr;
		cp->dport=info->sport;
		tcpha_be_conn_hash(cp);
		tcpha_be_set_state_timeout(cp,TCPHA_BE_S_ESTABLISHED);
		DBG("hashed connection:from:%u.%u.%u.%u:%d,to:%u.%u.%u.%u:%d\n",
				NIPQUAD(cp->caddr),ntohs(cp->cport),
				NIPQUAD(cp->daddr),ntohs(cp->dport));
		goto again;
	}
	else
		goto out_error;
	out:
	return 0;
	out_error:
	return -1;
}

void
tcpha_be_conn_release(struct tcpha_be_conn *conn)
{
	/* close the connetion with FE */
	if (conn->fsock) {
		DBG("close control connection:%d\n",conn->magic);
		sock_release(conn->fsock);
	}
	kmem_cache_free(tcpha_be_conn_cachep, conn);
}
