/*
  *  
  *	see tcpha_be.c for infoes about TCPHA
  *
  */
  
#include "tcpha_be.h"
#include "tcpha_be_misc.h"
#include "tcpha_be_con.h"

typedef enum http_methods {
	METHOD_NONE,
	METHOD_GET,
	METHOD_HEAD,
	METHOD_POST,
	METHOD_PUT,
	NR_METHODS
} http_method_t;

typedef enum http_versions {
	NOTKNOWN,
	HTTP_1_0,
	HTTP_1_1
} http_version_t;

/* HTTP MIME header */
typedef struct http_mime_header_s {
	struct list_head cookie_list;
	int content_length;
	int transfer_encoding;
	int connection_close;
	char *sep;		/* THIS_STRING_SEPARATES */
	int cookie;		/* if there is cookie in the header */
	int set_cookie2;
	ulong session_id;
} http_mime_header_t;

typedef struct http_request_s {
	const char *message;
	unsigned int message_len;
	unsigned int parsed_len;

	/* request method */
	int method;
	const char *method_str;

	/* request URI */
	const char *uri_str;
	unsigned int uri_len;
	unsigned int method_len;

	/* http version */
	int version;
	const char *version_str;
	unsigned int version_len;

	/* MIME header */
	http_mime_header_t mime;
} http_request_t;

#define GOTO_INCOMPLETE							\
do {									\
	ERR( "incomplete at %s:%d.\n", __FILE__, __LINE__);	\
	goto incomplete_message;					\
} while (0)

#define get_c(ptr,eob)				\
({						\
	if (ptr >= eob)				\
		GOTO_INCOMPLETE;		\
	*((ptr)++);				\
})

#define GOTO_ERROR							\
do {									\
	ERR("parse error at %s:%d.\n", __FILE__, __LINE__);	\
	goto error;							\
} while (0)

#define PARSE_TOKEN(ptr,str,eob)		\
({						\
	int __ret;				\
						\
	if (ptr+sizeof(str)-1 > eob) {		\
		GOTO_INCOMPLETE;		\
	}					\
						\
	if (memcmp(ptr, str, sizeof(str)-1))	\
		__ret = 0;			\
	else {					\
		ptr += sizeof(str)-1;		\
		__ret = 1;			\
	}					\
	__ret;					\
})

#define PARSE_OK		0
#define PARSE_INCOMPLETE	1
#define PARSE_ERROR		2

#define CR	13
#define LF	10
#define SP	' '

#define isspace(c)	((c) == ' ')

int
parse_http_header(const char *buffer, size_t buflen, http_request_t * req)
{
	const char *curr, *eob;
	char c;

	/* parse only the first header if multiple headers are present */

	curr = buffer;
	eob = buffer + buflen;

	req->message = buffer;
	req->message_len = buflen;

	/*
	 * RFC 2616, 5.1:
	 *      Request-Line = Method SP Request-URI SP HTTP-Version CRLF
	 */
	switch (get_c(curr, eob)) {
	case 'G':
		if (PARSE_TOKEN(curr, "ET ", eob)) {
			req->method = METHOD_GET;
			break;
		}
		GOTO_ERROR;

	case 'H':
		if (PARSE_TOKEN(curr, "EAD ", eob)) {
			req->method = METHOD_HEAD;
			break;
		}
		GOTO_ERROR;

	case 'P':
		if (PARSE_TOKEN(curr, "OST ", eob)) {
			req->method = METHOD_POST;
			break;
		}
		if (PARSE_TOKEN(curr, "UT ", eob)) {
			req->method = METHOD_PUT;
			break;
		}
		GOTO_ERROR;

	default:
		GOTO_ERROR;
	}
	req->method_str = buffer;
	req->method_len = curr - buffer - 1;

	req->uri_str = curr;
	while (1) {
		c = get_c(curr, eob);
		if (isspace(c))
			break;
	}
	req->uri_len = curr - req->uri_str - 1;

	req->version_str = curr;
	if (PARSE_TOKEN(curr, "HTTP/1.", eob)) {
		switch (get_c(curr, eob)) {
		case '0':
			req->version = HTTP_1_0;
			break;
		case '1':
			req->version = HTTP_1_1;
			break;
		default:
			GOTO_ERROR;
		}
	} else
		GOTO_ERROR;

	return PARSE_OK;

      incomplete_message:
	return PARSE_INCOMPLETE;

      error:
	return PARSE_ERROR;
}

static struct tcpha_be *
tcpha_be_http_matchrule(struct tcpha_be_service *svc, http_request_t *req)
{
	struct list_head *l;
	rule_t *rule=NULL;
	be_t *entry=NULL;
	struct tcpha_be *be = NULL;
	char tem=*(req->uri_str+req->uri_len);

	*(char *)(req->uri_str+req->uri_len)='\0';
	DBG("uri:%s\n",req->uri_str);
	list_for_each(l, &svc->rule_list) {
		rule = list_entry(l, rule_t, list);
        	if (!regexec(&rule->rx, req->uri_str,0, NULL, 0)) {
			break;
        	}
		rule=NULL;
	}
	*(char *)(req->uri_str+req->uri_len)=tem;
	if (!rule)
		goto out;
	spin_lock(&rule->__be_list_lock);
	list_for_each(l, &rule->be_list) {
		entry = list_entry(l, be_t, list);
		if (entry->be&&entry->be->state==TCPHA_BE_S_AVAIL)
			break;
		entry=NULL;
	}
	if (!entry) {
		spin_unlock(&rule->__be_list_lock);
		goto out;
	}
	list_del(&entry->list);
	list_add_tail(&entry->list,&rule->be_list);
	be=entry->be;
	spin_unlock(&rule->__be_list_lock);
	DBG("choose server:%u.%u.%u.%u\n",NIPQUAD(be->addr));
	out:
	return be;
}

struct tcpha_be *
tcpha_be_http_schedule(char *buf,int len,struct tcpha_be_service *svc)
{
	http_request_t req;
	
       if (parse_http_header(buf, len, &req)) {
	   	DBG("parse http packet error\n");
		return NULL;
       }
	return tcpha_be_http_matchrule(svc,&req);
}
