/*
 * TCPHA       An implementation of the TCP Virtual Server daemon inside
 *              kernel for the LINUX operating system.Different from KTCPVS,
 *		  TCPHA implements kenel layer-7 switch based on TCP Handoff.
 *		  TCPHA can be used to build a high-performance and highly available 
 *		   server based on a cluster of Linux servers(currently only 
 *              support BE which installs Linux,but i think it is NOT difficult to let Windows 
 *		   support TCP Handoff:),we will explore)
 *
 * Authors:     Li Wang  <dragonfly@linux-vs.org>
 *
 *              This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU General Public License
 *              as published by the Free Software Foundation; either version
 *              2 of the License, or (at your option) any later version.
 *
 */
 
#define __KERNEL_SYSCALLS__	/*  for waitpid */

#include "tcpha_fe.h"
#include "tcpha_fe_misc.h"
#include "tcpha_fe_file.h"
#include "tcpha_fe_hash.h"
#include "tcpha_fe_con.h"

enum {
	SERVER_DEAD = 0,
	SERVER_STARTING,
	SERVER_READY,
	SERVER_BUSY
};
	
#ifndef MAX_SPAWN_RATE
#define MAX_SPAWN_RATE	32
#endif

struct tcpha_fe_workshop {
	struct tcpha_fe_worker workers[TCPHA_FE_MAX_DAEMONS];
	int max_daemons_limit;
	int idle_spawn_rate;
	unsigned long last_modified;	/* last time of add/killing child */
};

#define NET_KTCPHAFE			20
#define NET_KTCPHAFE_UNLOAD	2

static int errno;
atomic_t tcphafe_adv_worker_count;
int advwkrpid;
int tcphafe_unload = 0;

static struct ctl_table_header *ktcphafe_table_header;

static ctl_table ktcphafe_table[] = {
	{NET_KTCPHAFE_UNLOAD, "unload", &tcphafe_unload,
	 sizeof(int), 0644, NULL, &proc_dointvec},
	{0}
};

static ctl_table ktcphafe_dir_table[] = {
	{NET_KTCPHAFE, "ktcphafe", NULL, 0, 0555, ktcphafe_table},
	{0}
};

static ctl_table ktcphafe_root_table[] = {
	{CTL_NET, "net", NULL, 0, 0555, ktcphafe_dir_table},
	{0}
};

static inline void
update_worker_status(struct tcpha_fe_worker *worker, __u16 status)
{
	worker->status = status;
}

static int worker_daemon(void *);
static int adv_worker_daemon(void *);

static inline void
hire_worker(struct tcpha_fe_workshop *tbl,
	   int slot, struct tcpha_fe_service *svc)
{
	if (slot + 1 > tbl->max_daemons_limit)
		tbl->max_daemons_limit = slot + 1;
	tbl->last_modified = jiffies;
	tbl->workers[slot].svc = svc;
	if (kernel_thread(worker_daemon, &tbl->workers[slot],
		CLONE_VM | CLONE_FS | CLONE_FILES) < 0)
		ERR("spawn child failed\n");
}

static inline void
hire_adv_worker(struct tcpha_fe_service *svc)
{
	if (!atomic_read(&tcphafe_adv_worker_count)) {
		if (kernel_thread(adv_worker_daemon, svc ,
			  	CLONE_VM | CLONE_FS | CLONE_FILES) < 0)
			ERR("spawn child failed\n");
		else
			atomic_inc(&tcphafe_adv_worker_count);
	}
}

static inline void
fire_worker(struct tcpha_fe_workshop *tbl, int slot,int cmd)
{
	kill_proc(tbl->workers[slot].pid,cmd, 1);
	tbl->last_modified = jiffies;
}

static inline void
tell_workers_offduty(struct tcpha_fe_workshop *tbl)
{
	int i;

	for (i=0;i<tbl->max_daemons_limit;i++) {
		if (tbl->workers[i].status!=SERVER_DEAD) {
			fire_worker(tbl,i,SIGSTOP);
		}
	}
	if (atomic_read(&tcphafe_adv_worker_count))
		kill_proc(advwkrpid,SIGSTOP,1);
}

static inline void
workshop_maintenance(struct tcpha_fe_workshop *tbl,
		       struct tcpha_fe_service *svc)
{
	int i;
	int free_slots[MAX_SPAWN_RATE];
	int free_length = 0;
	int to_kill = -1;
	int idle_count = 0;
	int last_non_dead = -1;

	hire_adv_worker(svc);
	for (i = 0; i < svc->conf.maxClients; i++) {
		int status;

		if (i >= tbl->max_daemons_limit
		    && free_length == tbl->idle_spawn_rate) break;
		status = tbl->workers[i].status;
		switch (status) {
		case SERVER_DEAD:
			if (free_length < tbl->idle_spawn_rate) {
				free_slots[free_length] = i;
				free_length++;
			}
			break;
		case SERVER_STARTING:
			idle_count++;
			last_non_dead = i;
			break;
		case SERVER_READY:
			idle_count++;
			to_kill = i;
			last_non_dead = i;
			break;
		case SERVER_BUSY:
			last_non_dead = i;
			break;
		}
	}
	tbl->max_daemons_limit = last_non_dead + 1;

	if (idle_count > svc->conf.maxSpareServers) {
		/* fire one worker each time */
		fire_worker(tbl, to_kill,SIGKILL);
		tbl->idle_spawn_rate = 1;
	} else if (idle_count < svc->conf.minSpareServers) {
		if (free_length) {
			if (tbl->idle_spawn_rate > 8 && net_ratelimit())
				INFO
				    ("Server seems busy, you may "
				     "need to increase StartServers, "
				     "or Min/MaxSpareServers\n");
			/* spawn a batch of children */
			for (i = 0; i < free_length; i++)
				hire_worker(tbl, free_slots[i], svc);

			if (tbl->idle_spawn_rate < MAX_SPAWN_RATE)
				tbl->idle_spawn_rate *= 2;
		} else if (net_ratelimit())
			INFO
			    ("Server reached MaxClients setting, "
			     "consider raising the MaxClients "
			     "setting\n");
	} else {
		/* if the number of spare servers remains in the interval
		   (minSpareServers, maxSpareServers] and the time of
		   last modified is larger than ten minutes, we try to
		   kill one spare child in order to release some resource. */
		if (idle_count > svc->conf.minSpareServers
		    && jiffies - tbl->last_modified > 600 * HZ)
			fire_worker(tbl, to_kill,SIGKILL);
		tbl->idle_spawn_rate = 1;
	}
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
static inline void
__init_prc_seg(void)
{
	spin_lock_irq(&current->sigmask_lock);
	siginitsetinv(&current->blocked,
		      sigmask(SIGKILL) | sigmask(SIGSTOP));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
}
#else
static inline void
__init_prc_seg(void)
{
	siginitsetinv(&current->blocked,
		      sigmask(SIGKILL) | sigmask(SIGSTOP));
	recalc_sigpending();
}
#endif
#define init_prc_seg() __init_prc_seg()

static int
worker_daemon(void *__wkr)
{
	struct tcpha_fe_conn *conn;
	struct socket *sock;
	int ret = 0;
	char *Buffer;
	size_t BufLen;
	struct tcpha_fe_worker *worker = (struct tcpha_fe_worker *) __wkr;
	struct tcpha_fe_service *svc = worker->svc;

	/* DECLARE_WAIT_QUEUE_HEAD(queue); */
	DECLARE_WAITQUEUE(wait, current);

	atomic_inc(&svc->workercount);
	worker->pid = current->pid;
	update_worker_status(worker, SERVER_STARTING);

	sprintf(current->comm, "tcpha_fe worker");
	lock_kernel();
	daemonize();

	/* Block all signals except SIGKILL and SIGSTOP */
	init_prc_seg();
	
	sock = svc->mainsock;
	if (sock == NULL) {
		ERR("socket for accept is NULL\n");
		ret = -1;
		goto out;
	}

	Buffer = (char *) get_free_page(GFP_KERNEL);
	if (Buffer == NULL) {
		ret = -ENOMEM;
		goto out;
	}
	BufLen = PAGE_SIZE;

	while (svc->stop == 0 && tcphafe_unload == 0) {
		if (signal_pending(current)) {
			DBG("worker (pid=%d): signal received\n",
				   current->pid);
			break;
		}

		update_worker_status(worker, SERVER_READY);
		if (sock->sk->tp_pinfo.af_tcp.accept_queue == NULL) {
			/* interruptible_sleep_on_timeout(&queue, HZ); */
			add_wait_queue(sock->sk->sleep, &wait);
			__set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ);
			__set_current_state(TASK_RUNNING);
			remove_wait_queue(sock->sk->sleep, &wait);
			continue;
		}

		/* create tcpha_fe_conn object */
		conn = tcpha_fe_conn_create(sock,svc, Buffer, BufLen);
		if (!conn)
			break;
		
		/* Do the actual accept */
		ret = sock->ops->accept(sock, conn->csock, O_NONBLOCK);
		if (ret < 0) {
			tcpha_fe_conn_release_now(conn);
			continue;
		}

		update_worker_status(worker, SERVER_BUSY);
		conn->state=TCPHA_FE_S_ESTABLISHED;
		
		/* Do the work */
		ret = tcpha_fe_conn_handle(conn, svc);
		if (ret < 0) {
			DBG_RL("Error handling connection\n");
			break;
		}
	}
	free_page((unsigned long) Buffer);
	
      out:
	update_worker_status(worker, SERVER_DEAD);
	atomic_dec(&svc->workercount);
	return ret;
}

static int
adv_worker_daemon(void *__svc)
{
	struct socket *sock,*csock;
	int ret = 0;
	struct tcpha_fe_info *Buffer;
	struct tcpha_fe_service *svc = (struct tcpha_fe_service *)__svc;

	/* DECLARE_WAIT_QUEUE_HEAD(queue); */
	DECLARE_WAITQUEUE(wait, current);

	atomic_inc(&svc->workercount);
	advwkrpid = current->pid;

	sprintf(current->comm, "tcpha_fe advwkr");
	lock_kernel();
	daemonize();

	/* Block all signals except SIGKILL and SIGSTOP */
	init_prc_seg();
	
	sock = svc->advsock;
	if (sock == NULL) {
		ERR("socket for accept is NULL\n");
		ret = -1;
		goto out;
	}

	Buffer = (struct tcpha_fe_info *) kmalloc(sizeof(*Buffer),GFP_KERNEL);
	if (Buffer == NULL) {
		ret = -ENOMEM;
		goto out;
	}

	while (svc->stop == 0 && tcphafe_unload == 0) {
		if (signal_pending(current)) {
			DBG("worker (pid=%d): signal received\n",
				   current->pid);
			break;
		}

		if (sock->sk->tp_pinfo.af_tcp.accept_queue == NULL) {
			/* interruptible_sleep_on_timeout(&queue, HZ); */
			add_wait_queue(sock->sk->sleep, &wait);
			__set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ);
			__set_current_state(TASK_RUNNING);
			remove_wait_queue(sock->sk->sleep, &wait);
			continue;
		}

		csock = sock_alloc();
		if (!csock) {
			break;
		}
		csock->type = sock->type;
		csock->ops = sock->ops;
		
		/* Do the actual accept */
		ret = sock->ops->accept(sock, csock, O_NONBLOCK);
		if (ret < 0) {
			DBG("accept error\n");
			continue;
		}

		/* Do the work */
		ret = tcpha_fe_adv_conn_handle(svc,csock,Buffer);
		if (ret < 0) {
			DBG_RL("Error handling connection\n");
			break;
		}
	}
	kfree(Buffer);
	
      out:
	atomic_dec(&tcphafe_adv_worker_count);
	atomic_dec(&svc->workercount);
	return ret;
}

static int
supervisor_daemon(void *__service)
{
	int waitpid_result;
	int i;
	struct tcpha_fe_service *svc = (struct tcpha_fe_service *) __service;
	struct tcpha_fe_workshop *workshop = NULL;

	MOD_INC_USE_COUNT;
	
	DECLARE_WAIT_QUEUE_HEAD(WQ);

	sprintf(current->comm, "tcpha_fe master");
	lock_kernel();
	daemonize();

	/* Block all signals except SIGKILL and SIGSTOP */
	init_prc_seg();
	
	workshop = vmalloc(sizeof(*workshop));
	if (!workshop)
		goto out;

	/* Then start listening and spawn the daemons */
	if (SetupServerSocket(svc) < 0)
		goto out;

	atomic_set(&svc->running, 1);
	atomic_set(&svc->workercount, 0);
	atomic_set(&tcphafe_adv_worker_count,0);
	svc->stop = 0;

	memset(workshop, 0, sizeof(*workshop));
	workshop->idle_spawn_rate = 1;

	hire_adv_worker(svc);
	for (i = 0; i < svc->conf.startservers; i++)
		hire_worker(workshop, i, svc);

	/* Then wait for deactivation */
	while (svc->stop == 0 && !signal_pending(current)
	       && tcphafe_unload == 0) {
		interruptible_sleep_on_timeout(&WQ, HZ);

		/* dynamically keep enough thread to handle load */
		workshop_maintenance(workshop, svc);

		/* reap the zombie daemons */
		waitpid_result = waitpid(-1, NULL, __WCLONE | WNOHANG);
	}

	tell_workers_offduty(workshop);
	
	/* Wait for tcp_vs_child to stop, one second per iteration */
	while (atomic_read(&svc->workercount) > 0)
		interruptible_sleep_on_timeout(&WQ, HZ);

	/* reap the zombie daemons */
	waitpid_result = 1;
	while (waitpid_result > 0)
		waitpid_result = waitpid(-1, NULL, __WCLONE | WNOHANG);

	/* stop listening */
	ReleaseServerSocket(svc);

      out:
	if (workshop)
		vfree(workshop);
	svc->start = 0;
	atomic_set(&svc->running, 0);

	MOD_DEC_USE_COUNT;
	
	return 0;
}

extern struct nf_hook_ops tcpha_fe_in_ops;
extern struct nf_hook_ops tcpha_fe_out_ops;

void flush_control_conn(struct tcpha_be *be)
{
	struct list_head *l;
	struct tcpha_control_conn *conn;

	for (l = be->conn_list.next; l != &be->conn_list;) {
		conn = list_entry(l, struct tcpha_control_conn, list);
		l = conn->list.next;
		tcpha_fe_close_ctrlconn(conn);
	}
}

void flush_be_list(struct tcpha_fe_service *svc)
{
	struct list_head *l;
	struct tcpha_be *be;

	for (l = svc->be_list.next; l != &svc->be_list;) {
		be = list_entry(l, struct tcpha_be, list);
		l = be->list.next;
		list_del(&be->list);
		flush_control_conn(be);
		kfree(be);
	}
}

void flush_be_t(rule_t *rule)
{
	struct list_head *l;
	be_t *be;

	for (l = rule->be_list.next; l != &rule->be_list;) {
		be = list_entry(l, be_t, list);
		l = be->list.next;
		list_del(&be->list);
		kfree(be);
	}
}

void flush_rule_list(struct tcpha_fe_service *svc)
{
	struct list_head *l;
	rule_t *rule;

	for (l = svc->rule_list.next; l != &svc->rule_list;) {
		rule = list_entry(l, rule_t, list);
		l = rule->list.next;
		list_del(&rule->list);
		flush_be_t(rule);
		regfree(&rule->rx);
		kfree(rule);
	}
}

void flush_tcpha_service(struct tcpha_fe_service *svc)
{
	flush_be_list(svc);
	flush_rule_list(svc);
}

static int __init
tcpha_fe_init(void)
{
	int ret=0;

	ret=config_tcpha_fe(&tfs);
	if (ret < 0) {
		ERR("can't config tcpha service\n");
		goto out_error;
	}
	ret = tcpha_fe_connhash_init();
	if (ret < 0) {
		ERR("can't setup connection table.\n");
		goto out_error;
	}
	ret = nf_register_hook(&tcpha_fe_in_ops);
	if (ret < 0) {
		ERR("can't register in hook.\n");
		goto cleanup_conn;
	}
	ret = nf_register_hook(&tcpha_fe_out_ops);
	if (ret < 0) {
		ERR("can't register out hook.\n");
		goto cleanup_inops;
	}
	ktcphafe_table_header = register_sysctl_table(ktcphafe_root_table, 0);
	atomic_set(&tcpha_fe_conn_count,0);
	(void) kernel_thread(supervisor_daemon,&tfs, 0);
	INFO("loaded.\n");
	return 0;
  	cleanup_inops:
	nf_unregister_hook(&tcpha_fe_in_ops);
  	cleanup_conn:
	tcpha_fe_conn_cleanup();
	out_error:
	flush_tcpha_service(&tfs);
	return ret;
}


static void __exit
tcpha_fe_cleanup(void)
{
	nf_unregister_hook(&tcpha_fe_out_ops);
	nf_unregister_hook(&tcpha_fe_in_ops);
	unregister_sysctl_table(ktcphafe_table_header);
	tcpha_fe_conn_cleanup();
	flush_tcpha_service(&tfs);
	INFO("unloaded.\n");
}


module_init(tcpha_fe_init);
module_exit(tcpha_fe_cleanup);
MODULE_LICENSE("GPL");
