/*
  *  
  *	see tcpha_fe.c for infoes about TCPHA
  *
  */
  
#include "tcpha_fe.h"
#include "tcpha_fe_con.h"

#define icmp_id(icmph)          ((icmph->un).echo.id)

/* get connection state,only called from bh */
static inline int get_conn_state(struct tcpha_fe_conn *cp)
{
	int ret;
	
	read_lock(&cp->lock);
	ret=cp->state;
	read_unlock(&cp->lock);

	return ret;
}

/*
 *	Handle ICMP messages in the outside-to-inside direction (incoming)
 *	Find any that might be relevant, check against existing connections,
 *	forward to the right destination host if relevant.
 *	Currently handles error types - unreachable, quench, ttl exceeded.
 */
static int tcpha_fe_in_icmp(struct sk_buff **skb_p)
{
	struct sk_buff	*skb   = *skb_p;
	struct iphdr    *iph;
	struct icmphdr  *icmph;
	struct iphdr    *ciph;	/* The ip header contained within the ICMP */
	__u16	        *pptr;	/* port numbers from TCP/UDP contained header */
	unsigned short   len;
	unsigned short	clen, csize;
	struct tcpha_fe_conn *cp;

	if (skb_is_nonlinear(skb)) {
		if (skb_linearize(skb, GFP_ATOMIC) != 0)
			return NF_DROP;
	}

	iph = skb->nh.iph;
	ip_send_check(iph);
	icmph = (struct icmphdr *)((char *)iph + (iph->ihl << 2));
	len = ntohs(iph->tot_len) - (iph->ihl<<2);
	if (len < sizeof(struct icmphdr))
		return NF_DROP;

	DBG("icmp in (%d,%d) %u.%u.%u.%u -> %u.%u.%u.%u\n",
		  icmph->type, ntohs(icmp_id(icmph)),
		  NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));

	if ((icmph->type != ICMP_DEST_UNREACH) &&
	    (icmph->type != ICMP_SOURCE_QUENCH) &&
	    (icmph->type != ICMP_TIME_EXCEEDED))
		return NF_ACCEPT;

	/*
	 * If we get here we have an ICMP error of one of the above 3 types
	 * Now find the contained IP header
	 */
	clen = len - sizeof(struct icmphdr);
	if (clen < sizeof(struct iphdr))
		return NF_DROP;
	ciph = (struct iphdr *) (icmph + 1);
	csize = ciph->ihl << 2;
	if (clen < csize)
		return NF_DROP;

	/* We are only interested ICMPs generated from TCP packets */
	if (ciph->protocol != IPPROTO_TCP)
		return NF_ACCEPT;

	/* Skip non-first embedded TCP/UDP fragments */
	if (ciph->frag_off & __constant_htons(IP_OFFSET))
		return NF_ACCEPT;

	/* We need at least TCP/UDP ports here */
	if (clen < csize + sizeof(struct udphdr))
		return NF_DROP;

	/* Ensure the checksum is correct */
	if (ip_compute_csum((unsigned char *) icmph, len)) {
		/* Failed checksum! */
		ERR_RL("incoming ICMP: failed checksum from "
			     "%d.%d.%d.%d!\n", NIPQUAD(iph->saddr));
		return NF_DROP;
	}

	pptr = (__u16 *)&(((char *)ciph)[csize]);

	DBG("Handling incoming ICMP for "
		  "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
		  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
		  NIPQUAD(ciph->daddr), ntohs(pptr[1]));

	/* This is pretty much what tcpha_fe_conn_in_get() does,
	   except parameters are in the reverse order */
	cp = tcpha_fe_conn_in_get(ciph->daddr, pptr[1],
			       ciph->saddr, pptr[0]);
	if (!cp)
		return NF_ACCEPT;

	int ret,state;
	
	state=get_conn_state(cp);
	switch (state) {
	case TCPHA_FE_S_NONE:
	case TCPHA_FE_S_ESTABLISHED:
		ret=NF_ACCEPT;
	default:
		ret=NF_STOLEN;
	}
	if (ret==NF_STOLEN) {
		if (cp->packet_xmit)
			ret = cp->packet_xmit(skb, cp);
		else
			ret = NF_ACCEPT;
	}
	switch (state) {
		case TCPHA_FE_S_NONE:
		case TCPHA_FE_S_ESTABLISHED:
		case TCPHA_FE_S_HANDOFFING:
			__tcpha_fe_conn_put(cp);
			break;
		default:
			tcpha_fe_conn_put(cp);
			
	}
	return ret;
}

/*
 *	transport layer header checking
 */
static inline int tcpha_fe_header_check(struct sk_buff *skb, int proto, int ihl)
{
	int len;

	switch (proto) {
	case IPPROTO_TCP:
		len = ihl + sizeof(struct tcphdr);
		/* we don't care about TCP options */
		break;
	default:
		len = 0;
	}

	/* guarantee protocol header available in skb data area */
	if (!pskb_may_pull(skb, len))
		return -1;
	else
		return 0;
}

/*
 *	Check if it's for virtual services, look it up,
 *	and send it on its way...
 */
static unsigned int tcpha_fe_in(unsigned int hooknum,
			     struct sk_buff **skb_p,
			     const struct net_device *in,
			     const struct net_device *out,
			     int (*okfn)(struct sk_buff *))
{
	struct sk_buff	*skb = *skb_p;
	struct iphdr	*iph = skb->nh.iph;
	union tcpha_fe_tphdr h;
	struct tcpha_fe_conn *cp;
	int ihl;

	/*
	 *	Big tappo: only PACKET_HOST (nor loopback neither mcasts)
	 *	... don't know why 1st test DOES NOT include 2nd (?)
	 */
	if (skb->pkt_type != PACKET_HOST || skb->dev == &loopback_dev) {
		return NF_ACCEPT;
	}

	if (iph->protocol == IPPROTO_ICMP)
		return tcpha_fe_in_icmp(skb_p);

	/* let it go if other IP protocols */
	if (iph->protocol != IPPROTO_TCP)
		return NF_ACCEPT;

	/* make sure that protocol header available in skb data area,
	   note that skb data area may be reallocated. */
	ihl = iph->ihl << 2;
	if (tcpha_fe_header_check(skb, iph->protocol, ihl) == -1)
		return NF_DROP;
	iph = skb->nh.iph;
	h.raw = skb->h.raw=(char*) iph + ihl;
	
	/*
	 * Check if the packet belongs to an existing connection entry
	 */
	cp = tcpha_fe_conn_in_get(iph->saddr, h.portp[0],
			       iph->daddr, h.portp[1]);

	if (!cp)
		return NF_ACCEPT;

	int ret=NF_STOLEN,state;

	state=get_conn_state(cp);
	switch (state) {
		case TCPHA_FE_S_NONE:
		case TCPHA_FE_S_ESTABLISHED:
			ret=NF_ACCEPT;
			break;
		case TCPHA_FE_S_HANDOFFING:
			ret=NF_DROP;
			break;
		default:
			break;
	}
	if (ret!=NF_DROP&&(h.th->fin||h.th->rst))
		tcpha_fe_set_state_timeout(cp,TCPHA_FE_S_TIME_WAIT);
	if (ret==NF_STOLEN) {
		if (cp->packet_xmit)
			ret = cp->packet_xmit(skb, cp);
		else {
			DBG_RL("warning: packet_xmit is null");
			ret = NF_ACCEPT;
		}
	}
	switch (state) {
		case TCPHA_FE_S_NONE:
		case TCPHA_FE_S_ESTABLISHED:
		case TCPHA_FE_S_HANDOFFING:
			__tcpha_fe_conn_put(cp);
			break;
		default:
			tcpha_fe_conn_put(cp);
			
	}
	return ret;
}

/* make sure that */
static unsigned int tcpha_fe_out(unsigned int hooknum,
			      struct sk_buff **skb_p,
			      const struct net_device *in,
			      const struct net_device *out,
			      int (*okfn)(struct sk_buff *))
{
	struct sk_buff  *skb = *skb_p;
	struct iphdr	*iph;
	union tcpha_fe_tphdr h;
	struct tcpha_fe_conn *cp;
	int ihl;

	iph = skb->nh.iph;

	/* let it go if other IP protocols */
	if (iph->protocol != IPPROTO_TCP)
		return NF_ACCEPT;

	/* make sure that protocol header available in skb data area,
	   note that skb data area may be reallocated. */
	ihl = iph->ihl << 2;
	if (tcpha_fe_header_check(skb, iph->protocol, ihl) == -1)
		return NF_DROP;

	iph = skb->nh.iph;
	h.th = (struct tcphdr *) iph + ihl;

	/*
	 *	Check if the packet belongs to an old entry
	 */
	cp = tcpha_fe_conn_out_get(iph->saddr, h.portp[0],
				iph->daddr, h.portp[1]);
	if (!cp) {
		return NF_ACCEPT;
	}

	int ret=NF_STOLEN,state;

	state=get_conn_state(cp);
	switch (state) {
		case TCPHA_FE_S_NONE:
			ret=NF_ACCEPT;
			break;
		case TCPHA_FE_S_ESTABLISHED:
			if (!h.th->fin&&!h.th->rst)
				ret=NF_ACCEPT;
			break;
		default:
			break;
	}
	switch (state) {
		case TCPHA_FE_S_NONE:
		case TCPHA_FE_S_ESTABLISHED:
		case TCPHA_FE_S_HANDOFFING:
			__tcpha_fe_conn_put(cp);
			break;
		default:
			tcpha_fe_conn_put(cp);
			
	}
	return ret;
}

struct nf_hook_ops tcpha_fe_in_ops = {
	{ NULL, NULL },
	tcpha_fe_in, PF_INET, NF_IP_LOCAL_IN, 100
};

struct nf_hook_ops tcpha_fe_out_ops = {
	{ NULL, NULL },
	tcpha_fe_out, PF_INET, NF_IP_FORWARD, 100
};
