/*
  *  
  *	see tcpha_fe.c for infoes about TCPHA
  *
  */
  
#include "tcpha_fe.h"
#include "tcpha_fe_misc.h"
#include "tcpha_fe_con.h"
#include "tcpha_fe_hash.h"

#ifdef	CONFIG_TCPHA_FE_DEBUG
int tcpha_fe_connmagic=0;
int tcpha_fe_ctrlconnmagic=0;
#define tf_set_connmagic(x) 	x->magic=tcpha_fe_connmagic++
#define tf_set_ctrlconnmagic(x)	x->magic=tcpha_fe_ctrlconnmagic++;
#else	/* NO DEBUGGING at ALL */
#define tf_set_connmagic(x) do {} while (0)
#define tf_set_ctrlconnmagic(x)  do {} while (0)
#endif

static const char * state_name_table[TCPHA_FE_S_LAST+1] = {
	[TCPHA_FE_S_NONE]          =	"NONE",
	[TCPHA_FE_S_ESTABLISHED]	=	"ESTABLISHED",
	[TCPHA_FE_S_HANDOFFING]		=	"HANDOFFING",
	[TCPHA_FE_S_HANDOFFED]		=	"HANDOFFED",
	[TCPHA_FE_S_TIME_WAIT]	=	"TIME_WAIT",
	[TCPHA_FE_S_CLOSE]         =	"CLOSE",
	[TCPHA_FE_S_CLOSE_WAIT]	=	"CLOSE_WAIT",
	[TCPHA_FE_S_ICMP]          =	"ICMP",
	[TCPHA_FE_S_LAST]          =	"BUG!",
};

/*
 *	Timeout table[state]
 */
struct timeout_table def_timeout_table = {
	ATOMIC_INIT(0),	/* refcnt */
	0,		/* scale  */
	{
		[TCPHA_FE_S_NONE]          =	30*60*HZ,	   /* not used */
		[TCPHA_FE_S_ESTABLISHED]	=	15*60*HZ,  /* not used */
		[TCPHA_FE_S_HANDOFFING]		=	1*60*HZ,    /* not used */
		[TCPHA_FE_S_HANDOFFED]		=	15*60*HZ,
		[TCPHA_FE_S_TIME_WAIT]	=	2*60*HZ,
		[TCPHA_FE_S_CLOSE]         =	10*HZ,
		[TCPHA_FE_S_CLOSE_WAIT]	=	60*HZ,
		[TCPHA_FE_S_ICMP]          =	1*60*HZ,
		[TCPHA_FE_S_LAST]          =	2*HZ,
	},	/* timeout */
};

static struct timeout_table *tcpha_fe_timeout_table = &def_timeout_table;

static inline void
tcpha_fe_timeout_attach(struct tcpha_fe_conn *cp, struct timeout_table *vstim)
{
	atomic_inc(&vstim->refcnt);
	cp->timeout_table = vstim;
}

int tcpha_fe_set_state_timeout(struct tcpha_fe_conn *cp, int state)
{
	struct timeout_table *vstim = cp->timeout_table;

	/*
	 *	Use default timeout table if no specific for this entry
	 */
	write_lock_bh(&cp->lock);
	if (!vstim)
		vstim = &def_timeout_table;

	cp->timeout = vstim->timeout[cp->state=state];

	if (vstim->scale) {
		int scale = vstim->scale;

		if (scale<0)
			cp->timeout >>= -scale;
		else if (scale > 0)
			cp->timeout <<= scale;
	}
	write_unlock_bh(&cp->lock);
	mod_timer(&cp->timer, jiffies+cp->timeout);
	
	return state;
}

static inline void tcpha_fe_timeout_detach(struct tcpha_fe_conn *cp)
{
	struct timeout_table *vstim = cp->timeout_table;

	if (!vstim)
		return;
	cp->timeout_table = NULL;
	atomic_dec(&vstim->refcnt);
}

static void tcpha_fe_conn_expire(unsigned long data)
{
	struct tcpha_fe_conn *cp = (struct tcpha_fe_conn *)data;

	DBG("connection:%d timeout\n",cp->magic);
	if (cp->timeout_table)
		cp->timeout = cp->timeout_table->timeout[TCPHA_FE_S_TIME_WAIT];
	else
		cp->timeout = def_timeout_table.timeout[TCPHA_FE_S_TIME_WAIT];

	/*
	 *	hey, I'm using it
	 */
	atomic_inc(&cp->refcnt);

	/*
	 *	unhash it if it is hashed in the conn table
	 */
	tcpha_fe_conn_unhash(cp);

	/*
	 *	refcnt==1 implies I'm the only one referrer
	 */
	if (likely(atomic_read(&cp->refcnt) <= 1)) {
		/* make sure that there is no timer on it now */
		if (timer_pending(&cp->timer))
			del_timer(&cp->timer);

		tcpha_fe_timeout_detach(cp);
		atomic_dec(&tcpha_fe_conn_count);

		kmem_cache_free(tcpha_fe_conn_cachep, cp);
		return;
	}

	/* hash it back to the table */
	tcpha_fe_conn_hash(cp);

	DBG("delayed: refcnt-1=%d\n",
		  atomic_read(&cp->refcnt)-1);
	tcpha_fe_conn_put(cp);
}


void tcpha_fe_conn_expire_now(struct tcpha_fe_conn *cp)
{
	cp->timeout = 0;
	mod_timer(&cp->timer, jiffies);
	__tcpha_fe_conn_put(cp);
}

const char * tcpha_fe_state_name(int state)
{
	if (state >= TCPHA_FE_S_LAST)
		return "ERR!";
	return state_name_table[state] ? state_name_table[state] : "?";
}

typedef enum ha_ret {
	TCP_HA_OK=0,
	TCP_HA_NO
}ha_ret_t;

typedef struct tcp_ha_ret {
	__u32 magic;
	__u32 connmagic;
	ha_ret_t msg;
}tcp_ha_ret_t;

void tcpha_fe_bind_xmit(struct tcpha_fe_conn *);

struct tcpha_fe_conn *
tcpha_fe_conn_create(struct socket *sock, struct tcpha_fe_service *svc,char *buffer, size_t buflen)
{
	struct tcpha_fe_conn *cp;

	cp = kmem_cache_alloc(tcpha_fe_conn_cachep, GFP_KERNEL);
	if (!cp) {
		ERR_RL("create_conn no memory available\n");
		return NULL;
	}
	memset(cp, 0, sizeof(*cp));
	/* clone the socket */
	cp->csock = sock_alloc();
	if (!cp->csock) {
		kmem_cache_free(tcpha_fe_conn_cachep, cp);
		return NULL;
	}
	cp->flags|=TCPHA_FE_CONN_F_TUNNEL; /* now only support this method,but it is enough:) */
	tf_set_connmagic(cp);
	init_timer(&cp->timer);
	cp->timer.data     = (unsigned long)cp;
	cp->timer.function = tcpha_fe_conn_expire;
	tcpha_fe_timeout_attach(cp, tcpha_fe_timeout_table);
	rwlock_init(&cp->lock);
	INIT_LIST_HEAD(&cp->list);
	atomic_inc(&tcpha_fe_conn_count);

	/* Set its state and timeout */
	cp->state=TCPHA_FE_S_NONE;

	/* Bind its packet transmitter */
	tcpha_fe_bind_xmit(cp);
	/* used by me */
	atomic_set(&cp->refcnt, 1);
	
	cp->csock->type = sock->type;
	cp->csock->ops = sock->ops;

	cp->buffer = buffer;
	cp->buflen = buflen;

	cp->raddr=svc->conf.raddr;
	cp->vaddr=svc->conf.vaddr;
	cp->vport=svc->conf.port;
	
	return cp;
}

struct tcpha_control_conn *tcpha_fe_new_ctrlconn(struct tcpha_be *be)
{
	struct tcpha_control_conn *cp;
	struct socket *sock;
	
	cp=(struct tcpha_control_conn *)kmalloc(sizeof(*cp),GFP_KERNEL);
	if (!cp)
		goto out_error;
	sock=tcpha_fe_connect2be(be);
	if (!sock) {
		ERR("connect to BE:%u.%u.%u.%u:%d error\n",NIPQUAD(be->addr),ntohs(be->port));
		kfree(cp);
		goto out_error;
	}
	cp->sock=sock;
	tf_set_ctrlconnmagic(cp);
	cp->be=be;
	DBG("created control connection:%d with:%u.%u.%u.%u:%d\n",cp->magic,NIPQUAD(be->addr),ntohs(be->port)); 
	return cp;
	out_error:
	return NULL;
}

void tcpha_fe_close_ctrlconn(struct tcpha_control_conn *cp)
{
	if (cp->be)
		atomic_dec(&cp->be->conn_count);
	if (cp->sock)
		sock_release(cp->sock);
	DBG("closed expired control connection:%d by process:%d\n",cp->magic,current->pid);
	kfree(cp);
}

static inline void tcpha_fe_get_ctrlconn(struct tcpha_control_conn *cp)
{
	list_del(&cp->list);
}

static inline void tcpha_fe_put_ctrlconn(struct tcpha_control_conn *cp)
{
	DBG("process:%d put control connection:%d\n",current->pid,cp->magic); 
	if (atomic_read(&cp->be->conn_count)>tfs.conf.conn_perbe) {
		tcpha_fe_close_ctrlconn(cp);
		return;
	}
	spin_lock(&cp->be->__conn_list_lock);
	list_add(&cp->list,&cp->be->conn_list);
	spin_unlock(&cp->be->__conn_list_lock);
}

/*
  *    get a control connection avaliable to BE
  */ 
struct tcpha_control_conn * tcpha_fe_getcon2be(struct tcpha_be *be)
{
	struct list_head *l;
	struct tcpha_control_conn *conn=NULL;
	
	spin_lock(&be->__conn_list_lock);
	list_for_each(l, &be->conn_list) {
		conn = list_entry(l, struct tcpha_control_conn, list);
		if (conn->sock) {
			tcpha_fe_get_ctrlconn(conn);
			break;
		}
		conn=NULL;
	}
	if (!conn) {
		spin_unlock(&be->__conn_list_lock);
		conn=tcpha_fe_new_ctrlconn(be);
		if (conn) {
			atomic_inc(&be->conn_count);
			goto out;
		}
		if (atomic_read(&be->conn_count)==0) {
			be->state=TCPHA_BE_S_NOTAVAIL;
			goto out;
		}
	}
	spin_unlock(&be->__conn_list_lock);
	out:
	return conn;
}

/* get TCP connection infoes */
void 
tcpha_fe_conn_getinfo(struct tcpha_fe_conn *conn,struct tcpha_conn_info *info)
{
	struct sock *sk=conn->csock->sk;
	
	read_lock_irq(&sk->callback_lock);
	info->daddr=sk->daddr;
	info->saddr=sk->saddr;
	info->dport=sk->dport;
	info->sport=sk->sport;
	info->rcv_saddr=sk->rcv_saddr;
	info->af_tcp=sk->tp_pinfo.af_tcp;
	read_unlock_irq(&sk->callback_lock);
	info->connmagic=sk->tp_pinfo.af_tcp.rcv_nxt;
	info->magic=TCPHAMAGIC;
}

/* reset a TCP connction */
static void tcp_reset(struct sock *sk)
{
	/* We want the right error as BSD sees it (and indeed as we do). */
	switch (sk->state) {
		case TCP_SYN_SENT:
			sk->err = ECONNREFUSED;
			break;
		case TCP_CLOSE_WAIT:
			sk->err = EPIPE;
			break;
		case TCP_CLOSE:
			return;
		default:
			sk->err = ECONNRESET;
	}

	if (!sk->dead)
		sk->error_report(sk);

	tcp_done(sk);
}

/*
  * 	reset the connection with client,client don't konw:),called by kernel thread
  */
void
tcpha_fe_conn_reset(struct tcpha_fe_conn *conn)
{
	DBG("reset connection:%d,refcnt:%d\n",conn->magic,atomic_read(&conn->refcnt)); 
	tcp_reset(conn->csock->sk);
	if (conn->csock->file) {
		fput(conn->csock->file);
		conn->csock->file=NULL;
	}
	else {
		iput(conn->csock->inode);
	}
	conn->csock=NULL;
	/* it is none of my business since that ... */
	atomic_dec(&conn->refcnt);
}

void tcpha_fe_conn_redirect(struct tcpha_fe_conn *conn)
{
	tcpha_fe_conn_expire_now(conn);
}

/* flush the conneciton entry,called by kernel thread */
void
tcpha_fe_conn_release(struct tcpha_fe_conn *conn)
{
	/* close the connetion with client */
	if (conn->csock) {
		DBG("close connection:%d with client:%u.%u.%u.%u:%d\n",conn->magic,
				NIPQUAD(conn->caddr),ntohs(conn->cport));
		sock_release(conn->csock);
	}
	/* relay the connection entry to bh */
	tcpha_fe_conn_expire_now(conn);
}

void tcpha_fe_conn_release_now(struct tcpha_fe_conn *conn)
{
	/* close the connetion with client */
	if (conn->csock) {
		DBG("close connection:%d with client:%u.%u.%u.%u:%d\n",conn->magic,
				NIPQUAD(conn->caddr),ntohs(conn->cport));
		sock_release(conn->csock);
	}
	tcpha_fe_timeout_detach(conn);
	atomic_dec(&tcpha_fe_conn_count);
	kfree(conn);
}

static inline int tcpha_fe_waitforpkt(struct socket *sock,long timeout)
{
	if (skb_queue_empty(&(sock->sk->receive_queue))) {
		if (!interruptible_sleep_on_timeout(&sock->wait, timeout)) {
			DBG("connection time out\n");
			goto out_error;
		}
		if (skb_queue_empty(&(sock->sk->receive_queue))) {
			DBG("connection closed by peer or received signal by process:%d\n",current->pid);
			goto out_error;
		}
	}
	return 0;
	out_error:
	return -1;
}

int 
tcpha_fe_adv_conn_handle(struct tcpha_fe_service *svc,struct socket *csock,struct tcpha_fe_info *buf)
{
	int len;
	tcp_ha_ret_t retmsg;
	int ret=-1;
	struct tcpha_fe_conn *cp;
	
	if (csock->sk->state != TCP_ESTABLISHED) {
		if (csock->sk->state == TCP_CLOSE_WAIT) {
			ret=0;
			goto out;
		}
		DBG_RL("Error connection not established (state %d)\n",
			   csock->sk->state);
		goto out;
	}

	if (tcpha_fe_waitforpkt(csock,15*HZ))
		goto out;
	/* fixme: what if the request overlap this receiving buffer */
	len = tcpha_fe_recvbuffer(csock,(char *)buf, sizeof(*buf), 0);
	if (len!=sizeof(*buf)||buf->magic!=TCPHAMAGIC) {
		DBG_RL("error reading request from BE\n");
		goto out;
	}
	cp=tcpha_fe_conn_in_get(buf->caddr,buf->cport,buf->daddr,buf->dport);
	if (!cp) {
		DBG_RL("cp not found\n");
		goto out;
	}
	write_lock_bh(&cp->lock);
	cp->daddr=buf->ndaddr;
	write_unlock_bh(&cp->lock);
	__tcpha_fe_conn_put(cp);
	retmsg.magic=TCPHAMAGIC;
	retmsg.msg=TCP_HA_OK;
	if (tcpha_fe_sendbuffer(csock,(char *)&retmsg,sizeof(retmsg),0)!=sizeof(retmsg)) {
		DBG_RL("send handoff response error\n");
		goto out;
	}
	DBG_RL("changed daddr to:%u.%u.%u.%u\n",NIPQUAD(buf->ndaddr));
	ret=0;
	out:
	sock_release(csock);
	return ret;
}

struct tcpha_be *
tcpha_fe_http_schedule(struct tcpha_fe_conn *, struct tcpha_fe_service *,int *);
int redirect_to_local(struct tcpha_fe_conn *, __u32, __u16);

int
tcpha_fe_conn_handle(struct tcpha_fe_conn *conn, struct tcpha_fe_service *svc)
{
	struct socket *csock, *dsock;
	struct tcpha_conn_info *info=(struct tcpha_conn_info *)conn->buffer;
	int len;
	tcp_ha_ret_t ret;
	struct tcpha_be *be;
	struct tcpha_control_conn *be_conn;
	
	csock = conn->csock;
	if (csock->sk->state != TCP_ESTABLISHED) {
		if (csock->sk->state == TCP_CLOSE_WAIT) {
			tcpha_fe_conn_release_now(conn);
			return 0;
		}
		DBG_RL("Error connection not established (state %d)\n",
			   csock->sk->state);
		tcpha_fe_conn_release_now(conn);
		return -1;
	}

	conn->caddr=csock->sk->daddr;
	conn->cport=csock->sk->dport;
	tcpha_fe_conn_hash(conn);

	DBG("created connection:%d with %u.%u.%u.%u:%d by:%d\n",conn->magic,
			NIPQUAD(conn->caddr),ntohs(conn->cport),current->pid);
	if (!(be=tcpha_fe_http_schedule(conn, svc,&len))) {
		if (svc->conf.redirect_port) {
			redirect_to_local(conn, svc->conf.redirect_addr,
					  svc->conf.redirect_port);
			tcpha_fe_conn_redirect(conn);
			return 0;
		}
		ERR_RL("no destination available\n");
		goto out_error;
	}

	be_conn = tcpha_fe_getcon2be(be);
	if (!be_conn) {
		ERR_RL("The destination is not available\n");
		if (svc->conf.redirect_port) {
			redirect_to_local(conn, svc->conf.redirect_addr,
					  svc->conf.redirect_port);
			tcpha_fe_conn_redirect(conn);
			return 0;
		}
		goto out_error;
	}
	DBG("process:%d grabed control connection:%d\n",current->pid,be_conn->magic);
	dsock = be_conn->sock;
	if (!dsock) {
		goto out_close_error;
	}
	conn->daddr=be->addr;
	conn->dport=be->port;
	tcpha_fe_conn_getinfo(conn,info);
	len+=TCPHACONNINFOSIZE;
	
	if (tcpha_fe_sendbuffer(dsock, conn->buffer, len, 0) != len) {
		DBG_RL("Error HTTP sending buffer\n");
		goto out_close_error;
	}

	if (tcpha_fe_waitforpkt(dsock,15*HZ))
		goto out_close_error;
	len = tcpha_fe_recvbuffer(dsock,(char *)&ret,sizeof(ret),0);
	if (len!=sizeof(ret)||ret.connmagic!=info->connmagic||ret.magic!=info->magic) {
		DBG_RL("error reading handoff response from BE\n");
		goto out_close_error;
	} 
	if (ret.msg!=TCP_HA_OK) {
		DBG_RL("handoff request refused by BE\n");
		goto out_close_error;
	}
	tcpha_fe_conn_reset(conn);
	tcpha_fe_set_state_timeout(conn, TCPHA_FE_S_HANDOFFED);
	tcpha_fe_put_ctrlconn(be_conn);
	return 0;

	out_close_error:
	tcpha_fe_close_ctrlconn(be_conn);
	out_error:
	tcpha_fe_conn_release(conn);
	return -1;
}
