/*
  *  
  *	see tcpha_fe.c for infoes about TCPHA
  *
  */
  
#include "tcpha_fe.h"
#include "tcpha_jhash.h"
#include "tcpha_fe_con.h"

/*
 *  Fine locking granularity for big connection hash table
 */
#define CT_LOCKARRAY_BITS  4
#define CT_LOCKARRAY_SIZE  (1<<CT_LOCKARRAY_BITS)
#define CT_LOCKARRAY_MASK  (CT_LOCKARRAY_SIZE-1)

struct tcpha_fe_aligned_lock
{
	rwlock_t	l;
} __attribute__((__aligned__(SMP_CACHE_BYTES)));

/* lock array for conn table */
struct tcpha_fe_aligned_lock
__tcpha_fe_conntbl_lock_array[CT_LOCKARRAY_SIZE] __cacheline_aligned;

static inline void ct_read_lock(unsigned key)
{
	read_lock(&__tcpha_fe_conntbl_lock_array[key&CT_LOCKARRAY_MASK].l);
}

static inline void ct_read_unlock(unsigned key)
{
	read_unlock(&__tcpha_fe_conntbl_lock_array[key&CT_LOCKARRAY_MASK].l);
}

static inline void ct_write_lock(unsigned key)
{
	write_lock(&__tcpha_fe_conntbl_lock_array[key&CT_LOCKARRAY_MASK].l);
}

static inline void ct_write_unlock(unsigned key)
{
	write_unlock(&__tcpha_fe_conntbl_lock_array[key&CT_LOCKARRAY_MASK].l);
}

static inline void ct_read_lock_bh(unsigned key)
{
	read_lock_bh(&__tcpha_fe_conntbl_lock_array[key&CT_LOCKARRAY_MASK].l);
}

static inline void ct_read_unlock_bh(unsigned key)
{
	read_unlock_bh(&__tcpha_fe_conntbl_lock_array[key&CT_LOCKARRAY_MASK].l);
}

static inline void ct_write_lock_bh(unsigned key)
{
	write_lock_bh(&__tcpha_fe_conntbl_lock_array[key&CT_LOCKARRAY_MASK].l);
}

static inline void ct_write_unlock_bh(unsigned key)
{
	write_unlock_bh(&__tcpha_fe_conntbl_lock_array[key&CT_LOCKARRAY_MASK].l);
}


/*
 *	Returns hash value for TCPHA_FE connection entry
 */
static unsigned
tcpha_fe_conn_hashkey(__u32 addr, __u16 port)
{
	return jhash_2words(addr, port, tcpha_fe_conn_rnd)
		& TCPHA_FE_CONN_TAB_MASK;
}


/*
 *	Hashes tcpha_fe_conn in tcpha_fe_conn_tab by proto,addr,port.
 *	returns bool success.
 */
int tcpha_fe_conn_hash(struct tcpha_fe_conn *cp)
{
	unsigned hash;

	if (cp->flags & TCPHA_FE_CONN_F_HASHED) {
		ERR("tcpha_fe_conn_hash(): request for already hashed, "
			  "called from %p\n", __builtin_return_address(0));
		return 0;
	}

	/* Hash by client address and port */
	hash = tcpha_fe_conn_hashkey(cp->caddr, cp->cport);

	ct_write_lock(hash);

	list_add(&cp->list, &tcpha_fe_conn_tab[hash]);
	cp->flags |= TCPHA_FE_CONN_F_HASHED;
	atomic_inc(&cp->refcnt);

	ct_write_unlock(hash);

	return 1;
}


/*
 *	UNhashes tcpha_fe_conn from tcpha_fe_conn_tab.
 *	returns bool success.
 */
int tcpha_fe_conn_unhash(struct tcpha_fe_conn *cp)
{
	unsigned hash;

	if (!(cp->flags & TCPHA_FE_CONN_F_HASHED)) {
		ERR("tcpha_fe_conn_unhash(): request for unhash flagged, "
			  "called from %p\n", __builtin_return_address(0));
		return 0;
	}

	/* unhash it and decrease its reference counter */
	hash = tcpha_fe_conn_hashkey(cp->caddr, cp->cport);
	ct_write_lock(hash);

	list_del(&cp->list);
	cp->flags &= ~TCPHA_FE_CONN_F_HASHED;
	atomic_dec(&cp->refcnt);

	ct_write_unlock(hash);

	return 1;
}

/*
 *  Gets tcpha_fe_conn associated with supplied parameters in the tcpha_fe_conn_tab.
 *  Called for pkts coming from OUTside-to-INside.
 *	s_addr, s_port: pkt source address (foreign host)
 *	d_addr, d_port: pkt dest address (load balancer:FE)
 */
static inline struct tcpha_fe_conn *__tcpha_fe_conn_in_get
(__u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port)
{
	unsigned hash;
	struct tcpha_fe_conn *cp;
	struct list_head *l,*e;

	hash = tcpha_fe_conn_hashkey(s_addr, s_port);
	l = &tcpha_fe_conn_tab[hash];

	ct_read_lock(hash);

	for (e=l->next; e!=l; e=e->next) {
		cp = list_entry(e, struct tcpha_fe_conn, list);
		if (s_addr==cp->caddr && s_port==cp->cport &&
		    d_port==cp->vport) {
			/* HIT */
			atomic_inc(&cp->refcnt);
			ct_read_unlock(hash);
			return cp;
		}
	}

	ct_read_unlock(hash);

	return NULL;
}

struct tcpha_fe_conn *tcpha_fe_conn_in_get
(__u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port)
{
	struct tcpha_fe_conn *cp;

	cp = __tcpha_fe_conn_in_get(s_addr, s_port, d_addr, d_port);

	return cp;
}

/*
 *  Gets tcpha_fe_conn associated with supplied parameters in the tcpha_fe_conn_tab.
 *  Called for pkts coming from inside-to-OUTside.
 *	s_addr, s_port: pkt source address (inside host)
 *	d_addr, d_port: pkt dest address (foreign host)
 */
struct tcpha_fe_conn *tcpha_fe_conn_out_get
(__u32 s_addr, __u16 s_port, __u32 d_addr, __u16 d_port)
{
	unsigned hash;
	struct tcpha_fe_conn *cp, *ret=NULL;
	struct list_head *l,*e;

	/*
	 *	Check for "full" addressed entries
	 */
	hash = tcpha_fe_conn_hashkey(d_addr, d_port);
	l = &tcpha_fe_conn_tab[hash];

	ct_read_lock(hash);

	for (e=l->next; e!=l; e=e->next) {
		cp = list_entry(e, struct tcpha_fe_conn, list);
		if (d_addr == cp->caddr && d_port == cp->cport &&
		    s_port == cp->dport && s_addr == cp->daddr) {
			/* HIT */
			atomic_inc(&cp->refcnt);
			ret = cp;
			break;
		}
	}

	ct_read_unlock(hash);

	return ret;
}

static int
tcpha_fe_conn_showinfo(char *buffer, char **start, off_t offset, int length)
{
	off_t pos=0;
	int idx, len=0;
	char temp[70];
	struct tcpha_fe_conn *cp;
	struct list_head *l, *e;

	pos = 128;
	if (pos > offset) {
		len += sprintf(buffer+len, "%-127s\n",
			       "       From                Dest            State       Expires");
	}

	for(idx = 0; idx < TCPHA_FE_CONN_TAB_SIZE; idx++) {
		/*
		 *	Lock is actually only need in next loop
		 *	we are called from uspace: must stop bh.
		 */
		ct_read_lock_bh(idx);

		l = &tcpha_fe_conn_tab[idx];
		for (e=l->next; e!=l; e=e->next) {
			cp = list_entry(e, struct tcpha_fe_conn,list);
			pos += 128;
			if (pos <= offset)
				continue;
			if (timer_pending(&cp->timer)) {
				sprintf(temp,
					"%u.%u.%u.%u:%d    %u.%u.%u.%u:%d   %-11s  %7lu",
					NIPQUAD(cp->caddr),ntohs(cp->cport),
					NIPQUAD(cp->daddr),ntohs(cp->dport),
					tcpha_fe_state_name(cp->state),
					(cp->timer.expires-jiffies)/HZ);
			}
			else {
			sprintf(temp,
				"%u.%u.%u.%u:%d    %u.%u.%u.%u:%d   %-11s  %-11s",
				NIPQUAD(cp->caddr),ntohs(cp->cport),
				NIPQUAD(cp->daddr),ntohs(cp->dport),
				tcpha_fe_state_name(cp->state),
				"not started");
			}
			len += sprintf(buffer+len, "%-127s\n", temp);
			if (pos >= offset+length) {
				ct_read_unlock_bh(idx);
				goto done;
			}
		}
		ct_read_unlock_bh(idx);
	}

  done:
	*start = buffer+len-(pos-offset);       /* Start of wanted data */
	len = pos-offset;
	if (len > length)
		len = length;
	if (len < 0)
		len = 0;
	return len;
}

/*
  * 	Initialize the hash tables and relevant locks
  */
int tcpha_fe_connhash_init(void)
{
	int idx;

	/*
	 * Allocate the connection hash table and initialize its list heads
	 */
	tcpha_fe_conn_tab = vmalloc(TCPHA_FE_CONN_TAB_SIZE*sizeof(struct list_head));
	if (!tcpha_fe_conn_tab)
		return -ENOMEM;

	INFO("Connection hash table configured "
		   "(size=%d, memory=%ldKbytes)\n",
		   TCPHA_FE_CONN_TAB_SIZE,
		   (long)(TCPHA_FE_CONN_TAB_SIZE*sizeof(struct list_head))/1024);

	for (idx = 0; idx < TCPHA_FE_CONN_TAB_SIZE; idx++) {
		INIT_LIST_HEAD(&tcpha_fe_conn_tab[idx]);
	}

	for (idx = 0; idx < CT_LOCKARRAY_SIZE; idx++)  {
		__tcpha_fe_conntbl_lock_array[idx].l = RW_LOCK_UNLOCKED;
	}

	/* Allocate tcpha_fe_conn slab cache */
	tcpha_fe_conn_cachep = kmem_cache_create("tcpha_fe_conn",
					      sizeof(struct tcpha_fe_conn), 0,
					      SLAB_HWCACHE_ALIGN, NULL, NULL);
	if (!tcpha_fe_conn_cachep) {
		vfree(tcpha_fe_conn_tab);
		return -ENOMEM;
	}

	proc_net_create("tcpha_fe_conn", 0, tcpha_fe_conn_showinfo);

	/* calculate the random value for connection hash */
	get_random_bytes(&tcpha_fe_conn_rnd, sizeof(tcpha_fe_conn_rnd));

	return 0;
}

/*
 *      Flush all the connection entries in the tcpha_fe_conn_tab
 */
void tcpha_fe_conn_flush(void)
{
	int idx;
	struct tcpha_fe_conn *cp;
	struct list_head *l,*e;

  flush_again:
	for (idx=0; idx<TCPHA_FE_CONN_TAB_SIZE; idx++) {
		/*
		 *  Lock is actually needed in this loop.
		 */
		ct_write_lock_bh(idx);

		l = &tcpha_fe_conn_tab[idx];
		for (e=l->next; e!=l; e=e->next) {
			cp = list_entry(e, struct tcpha_fe_conn,list);
			atomic_inc(&cp->refcnt);
			ct_write_unlock(idx);

			tcpha_fe_conn_expire_now(cp);
			ct_write_lock(idx);
		}
		ct_write_unlock_bh(idx);
	}

	/* the counter may be not NULL, because maybe some conn entries
	   are run by slow timer handler or unhashed but still referred */
	if (atomic_read(&tcpha_fe_conn_count) != 0) {
		DBG("tcpha_fe_conn_count:%d\n",atomic_read(&tcpha_fe_conn_count));
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ);
		goto flush_again;
	}
}

void tcpha_fe_conn_cleanup(void)
{
	/* flush all the connection entries first */
	tcpha_fe_conn_flush();
	
	/* Release the empty cache */
	kmem_cache_destroy(tcpha_fe_conn_cachep);
	proc_net_remove("tcpha_fe_conn");
	vfree(tcpha_fe_conn_tab);
}
